#!/bin/env python
# -*- coding: utf-8 -*-
##
# noise_model.py: Data model and interfaces for open systems simulation noise
#     models.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from __future__ import annotations
import html

## DESIGN NOTES ##

# The functions in this module may take dependencies on QuTiP and NumPy,
# while neither of those are dependencies of the qsharp package on the whole.
# To avoid those becoming hard dependencies, we do not import either package
# at the top of this module, but do so inside each function that uses those
# dependencies.
#
# This has some performance and code maintenance implications, but allows for
# being permissive with respect to dependencies.

## IMPORTS ##

from typing import Any, List, Optional, Tuple, Union, TYPE_CHECKING
import qsharp
import json
import dataclasses


## EXPORTS ##

__all__ = [
    "get_noise_model",
    "set_noise_model",
    "get_noise_model_by_name",
    "set_noise_model_by_name",
    "to_generator",
    "NoiseModel",

    # State data model
    "Stabilizer",

    # Process data model
    "SequenceProcess",
    "MixedPauliProcess",
    "ChpDecompositionProcess",
    "UnsupportedProcess",

    # Generator data model
    "ExplicitEigenvalueDecomposition",
    "UnsupportedGenerator",
    "GeneratorCoset",

    # Instrument data model
    "EffectsInstrument",
    "ZMeasInstrument",

    # CHP decomposition data model
    "Hadamard",
    "Cnot",
    "Phase",
    "AdjointPhase",

    # Common utility functions
    "t1_dissipation",
    "t2_dissipation",
    "depolarizing_process",
]

## TYPE DEFINITIONS ##

if TYPE_CHECKING:
    import qutip
    import numpy as np
    State = Union[qutip.Qobj, "Stabilizer"]
    Process = Union[qutip.Qobj, "ChpDecompositionProcess", "MixedPauliProcess", "SequenceProcess"]
    Instrument = Any # TODO
    Generator = Union[qutip.Qobj, "ExplicitEigenvalueDecomposition", "UnsupportedGenerator"]

## PUBLIC FUNCTIONS ##

def get_noise_model():
    """
    Returns the current noise model used in simulating Q# programs with the
    `.simulate_noise` method.
    """
    noise_model = convert_to_arrays(qsharp.client.get_noise_model())
    # Convert {"Mixed": ...} and so forth to qobj.
    return NoiseModel(**convert_to_qobjs(noise_model))

def get_noise_model_by_name(name: str):
    """
    Returns the built-in noise model with a given name.

    :param name: The name of the noise model to be returned (either `ideal`
        or `ideal_stabilizer`).
    """
    json_data = qsharp.client.get_noise_model_by_name(name)
    noise_model = convert_to_arrays(json_data)
    # Convert {"Mixed": ...} and so forth to qobj.
    return NoiseModel(**convert_to_qobjs(noise_model))

def set_noise_model(noise_model: NoiseModel):
    """
    Sets the current noise model used in simulating Q# programs with the
    `.simulate_noise` method.
    """
    json_data = dumps(convert_to_rust_style(_as_jobj(noise_model)))
    qsharp.client.set_noise_model(json_data)

def set_noise_model_by_name(name, **kwargs):
    """
    Sets the current noise model used in simulating Q# programs with the
    `.simulate_noise` method to a built-in noise model, given by name.

    Any additional keyword arguments are used to modify properties of the
        noise model before it is set.

    :param name: The name of the noise model to be set (either `ideal`
        or `ideal_stabilizer`).
    """
    qsharp.client.set_noise_model_by_name(name)
    if kwargs:
        model = get_noise_model()
        for name, value in kwargs.items():
            setattr(model, name, value)
        set_noise_model(model)

def to_generator(hamiltonian, *dissipators, pre=None, post=None) -> GeneratorCoset:
    """
    Given a Hamiltonian and possibly some dissipative terms, returns a
    generator representing evolution of density operators under that
    Hamiltonian.

    Note that only generators represented by normal matrices (that is, matrices
    that commute with their adjoints) are currently supported; a warning will
    result if a dissipator would result in non-normal evolution.

    For example, continuous-time generalized amplitude damping dissipators can result in
    non-normal generators in some cases.

    :param hamiltonian: The coherent portion of evolution to be represented.
    :param dissipators: Lindblad (aka "jump") operators representing
        continuous-time noise.
    :param pre: A quantum process applied before continuous-time noise.
    :param post: A quantum process applied following continuous-time noise.
    """
    import qutip as qt

    ident = qt.qeye(hamiltonian.dims[0][0])
    liouvillian = qt.tensor(ident, hamiltonian) - qt.tensor(hamiltonian.trans(), ident)

    generator = -1j * liouvillian

    for element in dissipators:
        if not isinstance(element, list):
            element = [element]

        for diss in element:
            li_dag_li = diss.dag() * diss
            generator += qt.tensor(diss.conj(), diss) - 0.5 * (
                qt.tensor(ident, li_dag_li) +
                qt.tensor(li_dag_li.trans(), ident)
            )

    normal_check = generator.dag() * generator - generator * generator.dag()
    if normal_check.norm() >= 1e-8:
        import warnings
        warnings.warn(f"Generator was not normal ({normal_check.norm()}), noise models including this generator may not be supported for use with simulate_noise method.")

    return GeneratorCoset(generator=generator, pre=pre, post=post)

def t2_dissipation(t2: float):
    """
    Represents dissipation from dephasing (a.k.a. T2) evolution.

    :param t2: The characteristic time T2 for the given dephasing dissipator.
    """
    import qutip as qt
    import numpy as np
    return [np.sqrt(1 / t2) * qt.sigmaz()]

def t1_dissipation(t1: float):
    """
    Represents dissipation from generalized amplitude damping (a.k.a. T1)
    evolution, with the maximally mixed state as its fixed point.

    :param t2: The characteristic time T1 for the given GAD dissipator.
    """
    import qutip as qt
    import numpy as np
    p_fixed = 0.5
    return [
        np.sqrt(p_fixed) * np.sqrt(1 / t1) * qt.sigmap(),
        np.sqrt(1 - p_fixed) * np.sqrt(1 / t1) * qt.sigmam(),
    ]

def depolarizing_process(p):
    import qutip as qt
    i = qt.to_super(qt.qeye(2))
    x = qt.to_super(qt.sigmax())
    y = qt.to_super(qt.sigmay())
    z = qt.to_super(qt.sigmaz())
    return p * i + (1 - p) / 4 * (i + x + y + z)

## PUBLIC DATA MODEL ##

@dataclasses.dataclass
class NoiseModel:
    initial_state: State
    cnot: Process
    i: Process
    s: Process
    s_adj: Process
    t: Process
    t_adj: Process
    h: Process
    x: Process
    y: Process
    z: Process
    z_meas: Instrument
    rx: GeneratorCoset
    ry: GeneratorCoset
    rz: GeneratorCoset

    def _as_jobj(self):
        def ensure_coset(o):
            return (
                o
                if isinstance(o, GeneratorCoset)
                else GeneratorCoset(generator=o, pre=None, post=None)
            )

        return {
            'initial_state': self.initial_state,
            'cnot': self.cnot,
            'i': self.i,
            's': self.s,
            's_adj': self.s_adj,
            't': self.t,
            't_adj': self.t_adj,
            'h': self.h,
            'x': self.x,
            'y': self.y,
            'z': self.z,
            'z_meas': self.z_meas,
            'rx': ensure_coset(self.rx)._as_jobj(),
            'ry': ensure_coset(self.ry)._as_jobj(),
            'rz': ensure_coset(self.rz)._as_jobj(),
        }

    def _repr_html_(self):
        def try_repr(value):
            return (
                "Unsupported" if isinstance(value, (UnsupportedGenerator, UnsupportedProcess)) else
                value._repr_html_() if hasattr(value, "_repr_html_") else
                value._repr_latex_() if hasattr(value, "_repr_latex_") else
                f"<tt>{html.escape(repr(value))}</tt>"
            )
        def gate_row(name):
            gate = getattr(self, name)
            text = try_repr(gate)
            return f"""
                <tr>
                    <td>{name}</td>
                    <td style="text-align: left" colspan="2">{text}</td>
                </tr>
            """
        def cont_gate_row(name):
            gate = getattr(self, name)
            rowspan = 1 + (1 if gate.pre is not None else 0) + (1 if gate.post is not None else 0)
            res = f"""
                <tr>
                    <td rowspan="{rowspan}">{name}</td>
                    <td style="text-align: left">Generator</td>
                    <td style="text-align: left">{try_repr(gate.generator)}</td>
                </tr>
            """
            if gate.pre is not None:
                res += f"""
                    <tr>
                        <td style="text-align: left">Pre</td>
                        <td style="text-align: left">{try_repr(gate.pre)}</td>
                    </tr>
                """
            if gate.post is not None:
                res += f"""
                    <tr>
                        <td style="text-align: left">Post</td>
                        <td style="text-align: left">{try_repr(gate.post)}</td>
                    </tr>
                """
            return res

        gates = "\n".join(
            gate_row(name)
            for name in (
                'i', 's', 's_adj', 't', 't_adj', 'h', 'x', 'y', 'z',
            )
        )
        cont_gates = "\n".join(
            cont_gate_row(name)
            for name in (
                'rx', 'ry', 'rz',
            )
        )
        return f"""
            <table>
                <caption>Noise model</caption>
                <thead>
                    <tr>
                        <th>Name</th>
                        <th style="text-align: left" colspan="2">Value</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>Initial state</td>
                        <td style="text-align: left" colspan="2">{try_repr(self.initial_state)}</td>
                    </tr>
                    {gates}
                    {cont_gates}
                    <tr>
                        <td>$Z$-measurement</td>
                        <td>{try_repr(self.z_meas)}</td>
                    </tr>
                </tbody>
            </table>
        """

@dataclasses.dataclass
class Stabilizer:
    n_qubits: int
    table: np.ndarray

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'Stabilizer': {
                    'n_qubits': self.n_qubits,
                    # We don't use arr_to_rust_style here, as that is designed
                    # for complex arrays. Rather, we depend on the default
                    # set in as_jobj, below.
                    'table': self.table
                }
            }
        }

@dataclasses.dataclass
class UnsupportedProcess():
    n_qubits: int

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': 'Unsupported'
        }

@dataclasses.dataclass
class ExplicitEigenvalueDecomposition():
    values: Any
    vectors: Any
    n_qubits: int

    @classmethod
    def _from_qobj(cls, qobj):
        import qutip as qt
        import numpy as np

        vals, vecs = qobj.eigenstates()
        # TODO: Double check n_qubits!
        # TODO: Truncate zero eigenvalues.
        return cls(
            vals,
            np.array([np.asarray(vec.data.todense())[:, 0] for vec in vecs]),
            1
        )

    def _repr_latex_(self):
        import qutip as qt
        res = qt.qzero([[2] * self.n_qubits] * 2)
        for val, vec in zip(self.values, self.vectors):
            proj = val * qt.Qobj(vec) * qt.Qobj(vec).dag()
            proj.dims = res.dims
            res += proj
        return res._repr_latex_().replace(
            "Quantum object",
            "Quantum object w/ precomputed eigendecomposition"
        )

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'ExplicitEigenvalueDecomposition': {
                    'values': arr_to_rust_style(self.values),
                    'vectors': arr_to_rust_style(self.vectors)
                }
            }
        }

@dataclasses.dataclass
class UnsupportedGenerator():
    n_qubits: int

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': 'Unsupported'
        }

@dataclasses.dataclass
class GeneratorCoset():
    generator: Union[Generator, "qutip.Qobj"]
    pre: Optional[Any]
    post: Optional[Any]

    def _as_jobj(self):
        try:
            import qutip as qt
        except:
            qt = None

        obj = {}

        if self.pre is not None:
            obj['pre'] = self.pre

        if self.post is not None:
            obj['post'] = self.post

        generator = self.generator

        if qt is not None and isinstance(generator, qt.Qobj):
            generator = ExplicitEigenvalueDecomposition._from_qobj(generator)

        obj['generator'] = _as_jobj(generator)

        return obj

@dataclasses.dataclass
class SequenceProcess():
    n_qubits: int
    processes: List[Any]

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                "Sequence": list(map(_as_jobj, self.processes))
            }
        }

@dataclasses.dataclass
class Hadamard():
    idx_target: int

    def __repr__(self):
        return f"Hadamard({self.idx_target})"

    def _as_jobj(self):
        return {
            'Hadamard': self.idx_target
        }

@dataclasses.dataclass
class Phase():
    idx_target: int

    def __repr__(self):
        return f"Phase({self.idx_target})"

    def _as_jobj(self):
        return {
            'Phase': self.idx_target
        }

@dataclasses.dataclass
class AdjointPhase():
    idx_target: int

    def __repr__(self):
        return f"AdjointPhase({self.idx_target})"

    def _as_jobj(self):
        return {
            'AdjointPhase': self.idx_target
        }

@dataclasses.dataclass
class Cnot():
    idx_control: int
    idx_target: int

    def __repr__(self):
        return f"Cnot({self.idx_control}, {self.idx_target})"

    def _as_jobj(self):
        return {
            'Cnot': [self.idx_control, self.idx_target]
        }

@dataclasses.dataclass
class ChpDecompositionProcess():
    n_qubits: int
    operations: List[Union[Hadamard, Cnot, Phase, AdjointPhase]]

    @classmethod
    def _from_jobj(cls, obj):
        return cls(n_qubits=obj['n_qubits'], operations=[
            Hadamard(step['Hadamard'])
            if 'Hadamard' in step else
            Phase(step['Phase'])
            if 'Phase' in step else
            AdjointPhase(step['AdjointPhase'])
            if 'AdjointPhase' in step else
            Cnot(*step['Cnot'])
            for step in obj['data']['ChpDecomposition']
        ])

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'ChpDecomposition': list(map(_as_jobj, self.operations))
            }
        }

@dataclasses.dataclass
class MixedPauliProcess():
    n_qubits: int
    operators: List[Tuple[float, Union[List[qsharp.Pauli], str]]]

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'MixedPauli': [
                    [
                        pr,
                        [
                            (qsharp.Pauli[p] if isinstance(p, str) else qsharp.Pauli(p)).name
                            for p in ops
                        ]
                    ]
                    for (pr, ops) in self.operators
                ]
            }
        }

@dataclasses.dataclass
class EffectsInstrument():
    effects: List[Process]

    def _as_jobj(self):
        return {
            'Effects': list(map(qobj_to_rust_style, self.effects))
        }

@dataclasses.dataclass
class ZMeasInstrument():
    pr_readout_error: float

    def _as_jobj(self):
        return {
            'ZMeasurement': {
                'pr_readout_error': self.pr_readout_error
            }
        }

## PRIVATE FUNCTIONS ##

literal_keys = frozenset(['ChpDecomposition'])

def _as_jobj(o, default=lambda x: x):
    import numpy as np
    if isinstance(o, np.ndarray):
        # Use Rust-style arrays.
        return {
            'v': 1,
            'dim': list(o.shape),
            'data': o.reshape((-1, )).tolist()
        }
    elif hasattr(o, '_as_jobj'):
        return o._as_jobj()

    return default(o)

class NoiseModelEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return _as_jobj(o, super().default)

def dumps(obj: Any) -> str:
    """
    Wraps json.dumps with a custom JSONEncoder class to cover types used in
    noise model serialization.
    """
    return json.dumps(
        obj=obj,
        cls=NoiseModelEncoder
    )

def is_rust_style_array(json_obj):
    return (
        isinstance(json_obj, dict) and
        'v' in json_obj and
        'data' in json_obj and
        'dim' in json_obj and
        json_obj['v'] == 1
    )

def rust_style_array_as_array(json_obj, as_complex : bool = True):
    import numpy as np
    dims = json_obj['dim'] + [2] if as_complex else json_obj['dim']
    arr = np.array(json_obj['data']).reshape(dims)
    if as_complex:
        arr = arr.astype(complex)
        return arr[..., 0] + 1j * arr[..., 1]
    else:
        return arr

def convert_to_arrays(json_obj, as_complex : bool = True):
    return (
        # Add on a trailing index of length 2.
        rust_style_array_as_array(json_obj, as_complex=as_complex)
        if is_rust_style_array(json_obj) else
        {
            key:
                value
                if key in literal_keys else

                convert_to_arrays(value, as_complex=key != 'table')
                if isinstance(value, dict) else

                [convert_to_arrays(element) for element in value]
                if isinstance(value, list) else

                value

            for key, value in json_obj.items()
        }
    )

def arr_to_qobj(arr):
    import qutip as qt
    import numpy as np
    return qt.Qobj(arr, dims=[[2] * int(np.log2(arr.shape[1]))] * 2)


def jobj_to_generator(json_obj) -> GeneratorCoset:
    generator = json_obj['generator']
    if 'data' in generator:
        if generator['data'] == "Unsupported":
            generator = UnsupportedGenerator(generator['n_qubits'])
        elif 'ExplicitEigenvalueDecomposition' in generator['data']:
            generator = ExplicitEigenvalueDecomposition(
                values=generator['data']['ExplicitEigenvalueDecomposition']['values'],
                vectors=generator['data']['ExplicitEigenvalueDecomposition']['vectors'],
                n_qubits=generator['n_qubits']
            )

    return GeneratorCoset(
        generator=generator,
        pre=convert_to_qobjs(json_obj['pre']) if 'pre' in json_obj else None,
        post=convert_to_qobjs(json_obj['post']) if 'post' in json_obj else None
    )


def convert_to_qobjs(json_obj):
    import qutip as qt
    return (
        arr_to_qobj(json_obj['data']['Mixed'])
        if 'data' in json_obj and 'Mixed' in json_obj['data'] else

        arr_to_qobj(json_obj['data']['Unitary'])
        if 'data' in json_obj and 'Unitary' in json_obj['data'] else

        jobj_to_generator(json_obj)
        if 'generator' in json_obj else

        Stabilizer(json_obj['n_qubits'], json_obj['data']['Stabilizer']['table'])
        if 'data' in json_obj and 'Stabilizer' in json_obj['data'] else

        ChpDecompositionProcess._from_jobj(json_obj)
        if 'data' in json_obj and 'ChpDecomposition' in json_obj['data'] else

        SequenceProcess(json_obj['n_qubits'], list(map(convert_to_qobjs, json_obj['data']['Sequence'])))
        if 'data' in json_obj and 'Sequence' in json_obj['data'] else

        # TODO: Hook up effects instrument, z instrument.
        EffectsInstrument(list(map(convert_to_qobjs, json_obj['Effects'])))
        if 'Effects' in json_obj else

        ZMeasInstrument(json_obj['ZMeasurement']['pr_readout_error'])
        if 'ZMeasurement' in json_obj else

        # Since we deserialized unsupported generators above, we know that
        # this unsupported must be a process.
        UnsupportedProcess(json_obj['n_qubits'])
        if 'data' in json_obj and json_obj['data'] == "Unsupported" else

        qt.kraus_to_super([arr_to_qobj(op) for op in json_obj['data']['KrausDecomposition']])
        if 'data' in json_obj and 'KrausDecomposition' in json_obj['data'] else

        {
            key:
                value
                if key in literal_keys else

                # Recurse if needed...
                convert_to_qobjs(value)
                if isinstance(value, dict) else

                [convert_to_qobjs(element) for element in value]
                if isinstance(value, list) else

                # Just return value if there's nothing else to do.
                value
            for key, value in json_obj.items()
        }
    )

def arr_to_rust_style(arr):
    import numpy as np
    return {
        'v': 1,
        'dim': list(arr.shape),
        'data': np.moveaxis(np.array([arr.real, arr.imag]), 0, -1).reshape((-1, 2)).tolist()
    }

def qobj_to_rust_style(qobj, expect_state=False):
    import qutip as qt
    import numpy as np
    data = None
    n_qubits = 1
    # Figure out what kind of qobj we have and convert accordingly.
    if qobj.type == 'oper':
        n_qubits = len(qobj.dims[0])
        data = {
            'Mixed' if expect_state else 'Unitary':
            arr_to_rust_style(qobj.data.todense())
        }
    elif qobj.type == 'super':
        n_qubits = len(qobj.dims[0][0])
        data = {
            'KrausDecomposition': arr_to_rust_style(
                np.array([
                    op.data.todense()
                    for op in qt.to_kraus(qobj)
                ])
            )
        }
    return {
        "n_qubits": n_qubits,
        "data": data
    }

def convert_to_rust_style(json_obj, expect_state=False):
    import qutip as qt
    return (
        qobj_to_rust_style(json_obj, expect_state=expect_state)
        if isinstance(json_obj, qt.Qobj) else

        list(map(convert_to_rust_style, json_obj))
        if isinstance(json_obj, list) else

        {
            key: convert_to_rust_style(value, expect_state=key == 'initial_state')
            for key, value in json_obj.items()
        }
        if isinstance(json_obj, dict) else

        json_obj
    )



#!/bin/env python
# -*- coding: utf-8 -*-
##
# utils.py: Utilities internal to the qsharp package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from dataclasses import dataclass
import logging
import warnings
logger = logging.getLogger(__name__)
from typing import Callable

## INTERNAL FUNCTIONS ##

def log_messages(data, action : Callable[[str], None] = logger.error):
    msgs = data['messages']
    for msg in msgs:
        action(msg)

@dataclass
class ImportFailure:
    cause: ImportError

_qutip_cache = None
def try_import_qutip(warn=False, optional=False):
    global _qutip_cache
    if _qutip_cache is None:
        try:
            import qutip
            _qutip_cache = qutip
        except ImportError as ex:
            _qutip_cache = ImportFailure(ex)

    if warn and isinstance(_qutip_cache, ImportFailure):
        warnings.warn(f"Failed to import QuTiP with error: {_qutip_cache.cause.msg}", source=_qutip_cache.cause)

    if optional and isinstance(_qutip_cache, ImportFailure):
        # suppress failure when optional
        return None

    return _qutip_cache



#!/bin/env python
# -*- coding: utf-8 -*-
##
# azure.py: enables using Q# quantum execution on Azure Quantum from Python.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

## IMPORTS ##

import qsharp
import json
from typing import List, Dict, Callable, Any, Union
from enum import Enum

## LOGGING ##

import logging
logger = logging.getLogger(__name__)

## EXPORTS ##

__all__ = [
    'connect',
    'target',
    'submit',
    'execute',
    'status',
    'output',
    'jobs',
    'AzureTarget',
    'AzureJob',
    'AzureError',
    'AzureResult'
]

## CLASSES ##

class AzureTarget(object):
    """
    Represents an instance of an Azure Quantum execution target for Q# job submission.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.id = data["id"]
        self.current_availability = data["current_availability"]
        self.average_queue_time = data["average_queue_time"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureTarget):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

class AzureJob(object):
    """
    Represents an instance of an Azure Quantum job.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.id = data["id"]
        self.name = data["name"]
        self.status = data["status"]
        self.uri = data["uri"]
        self.provider = data["provider"]
        self.target = data["target"]
        self.creation_time = data["creation_time"]
        self.begin_execution_time = data["begin_execution_time"]
        self.end_execution_time = data["end_execution_time"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureJob):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

class AzureResult(Dict):
    """
    Represents a the results of an Azure Quantum job.
    """
    def __init__(self, data: Dict, mime_bundle: Dict[str, Any] = None):
        super().__init__(data)
        self.mime_bundle = mime_bundle if mime_bundle is not None else {}

    def _repr_mimebundle_(self, include=None, exclude=None) -> Dict[str, Any]:
        return self.mime_bundle

class AzureError(Exception):
    """
    Contains error information resulting from an attempt to interact with Azure.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.error_code = data["error_code"]
        self.error_name = data["error_name"]
        self.error_description = data["error_description"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureError):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

## FUNCTIONS ##

def connect(**params) -> List[AzureTarget]:
    """
    Connects to an Azure Quantum workspace or displays current connection status.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.connect for more details.
    """
    result = qsharp.client._execute_magic(f"azure.connect", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return [AzureTarget(target) for target in result]

def target(name : str = '', **params) -> AzureTarget:
    """
    Sets or displays the active execution target for Q# job submission in an Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.target for more details.
    """
    result = qsharp.client._execute_magic(f"azure.target {name}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureTarget(result)

def target_capability(name : str = '', **params) -> Dict:
    """
    Sets or displays the active target capability for Q# job submission in an Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.target-capability for more details.
    """
    result = qsharp.client._execute_magic(f"azure.target-capability {name}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return result

def submit(op : qsharp.QSharpCallable, **params) -> AzureJob:
    """
    Submits a job to an Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.submit for more details.
    """
    result = qsharp.client._execute_callable_magic("azure.submit", op, raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureJob(result)

def execute(op : qsharp.QSharpCallable, **params) -> AzureResult:
    """
    Submits a job to an Azure Quantum workspace and waits for completion.
    Returns a dictionary containing the results of the job. This dictionary will usually
    have keys corresponding to the possible operation results and values indicating the
    percentage of shots that resulted in that key value, but may vary depending on the job's target.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.execute for more details.
    """
    (result, content) = qsharp.client._execute_callable_magic("azure.execute", op, raise_on_stderr=False, return_full_result=True, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureResult(result, content['data'])

def status(jobId : str = '', **params) -> AzureJob:
    """
    Displays status for a job in the current Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.status for more details.
    """
    result = qsharp.client._execute_magic(f"azure.status {jobId}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureJob(result)

def output(jobId : str = '', **params) -> AzureResult:
    """
    Displays results for a job in the current Azure Quantum workspace.
    Returns a dictionary containing the results of the job. This dictionary will usually
    have keys corresponding to the possible operation results and values indicating the
    percentage of shots that resulted in that key value, but may vary depending on the job's target.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.output for more details.
    """
    (result, content) = qsharp.client._execute_magic(f"azure.output {jobId}", raise_on_stderr=False, return_full_result=True, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureResult(result, content['data'])

def jobs(filter : str = '', count : int = 30, **params) -> List[AzureJob]:
    """
    Displays a list of jobs in the current Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.jobs for more details.
    """
    result = qsharp.client._execute_magic(f"azure.jobs \"{filter}\" count={count}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return [AzureJob(job) for job in result]



#!/bin/env python
# -*- coding: utf-8 -*-
##
# iqsharp.py: Client for the IQ# Jupyter kernel.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""
This module allow for using experimental features of the Quantum Development Kit,
including noisy simulators for Q# programs.
"""

import warnings

# Re-export noise model APIs from their new locations.
from qsharp.noise_model import *

def enable_noisy_simulation():
    warnings.warn(
        "The open systems simulation feature is no longer experimental, such that enable_noisy_simulation "
        "no longer needs to be called before using simulate_noise. This function will be removed in "
        "a future version of the qsharp package.",
        DeprecationWarning
    )



#!/bin/env python
# -*- coding: utf-8 -*-
##
# __init__.py: Root module for the qsharp package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

"""

"""

## IMPORTS ##

import sys
from contextlib import contextmanager
from typing import Any, List, Dict, Union
from collections import defaultdict
from distutils.version import LooseVersion

from qsharp.clients import _start_client
from qsharp.clients.iqsharp import IQSharpError
from qsharp.loader import QSharpCallable, QSharpModuleFinder
from qsharp.config import Config
from qsharp.packages import Packages
from qsharp.projects import Projects
from qsharp.types import Result, Pauli
from qsharp.utils import ImportFailure, try_import_qutip
try:
    from qsharp.version import __version__
except:
    __version__ = "<unknown>"

## EXPORTS ##

__all__ = [
    'compile', 'reload',
    'get_available_operations', 'get_available_operations_by_namespace',
    'get_workspace_operations',
    'config',
    'packages',
    'projects',
    'IQSharpError',
    'Result', 'Pauli'
]

from qsharp.noise_model import *
import qsharp.noise_model
__all__ += qsharp.noise_model.__all__

## FUNCTIONS ##

def compile(code : str) -> Union[None, QSharpCallable, List[QSharpCallable]]:
    """
    Given a string containing Q# source code, compiles it into the current
    workspace and returns one or more Q# callable objects that can be used to
    invoke the new code.

    :param code: A string containing Q# source code to be compiled.
    :returns: A list of callables compiled from `code`, or a callable if exactly
        one callable is found.
    """
    compiled = client.compile(code)
    if compiled is None:
        return None

    ops = [
        QSharpCallable(op, "snippets")
        for op in compiled
    ]
    if len(ops) == 1:
        return ops[0]
    else:
        return ops


def reload() -> None:
    """
    Reloads the current IQ# workspace, recompiling source files in the
    workspace.

    If the workspace fails to compile (e.g., because of a missing package),
    Q# compilation errors are raised as an exception.
    """
    client.reload()

def get_available_operations() -> List[str]:
    """
    Returns a list containing the names of all operations and functions defined
    in the current workspace, and that have been compiled dynamically from
    snippets.
    """
    return client.get_available_operations()

def get_workspace_operations() -> List[str]:
    """
    Returns a list containing the names of all operations and functions defined
    in the current workspace, excluding dynamically compiled snippets.
    """
    return client.get_workspace_operations()

def get_available_operations_by_namespace() -> Dict[str, List[str]]:
    """
    Returns a dictionary from namespaces to lists of operations and functions
    defined in the current workspace that are members of each namespace.
    """
    ops = get_available_operations()
    by_ns = defaultdict(list)

    for qualified_name in ops:
        idx_last_dot = qualified_name.rfind(".")
        ns_name = qualified_name[:idx_last_dot]
        op_name = qualified_name[idx_last_dot + 1:]

        by_ns[ns_name].append(op_name)

    return dict(by_ns.items())

def component_versions() -> Dict[str, LooseVersion]:
    """
    Returns a dictionary from components of the IQ# kernel to their
    versions.
    """
    versions = client.component_versions()
    # Add in the qsharp Python package itself.
    versions["qsharp"] = LooseVersion(__version__)
    # If any experimental features are enabled, report them here.
    if _experimental_versions is not None:
        versions['experimental'] = _experimental_versions
    return versions

@contextmanager
def capture_diagnostics(passthrough: bool = False, as_qobj: bool = False) -> List[Any]:
    """
    Returns a context manager that captures diagnostics output from running Q#
    programs into a list.

    For example, to capture `DumpMachine` calls from a Q# operation run on the
    full-state simulator:

    .. code-block:: qsharp

        namespace Sample {
            open Microsoft.Quantum.Intrinsic;
            open Microsoft.Quantum.Diagnostics;

            operation RunMain() : Unit {
                use q = Qubit();
                within {
                    H(q);
                } apply {
                    DumpMachine();
                }
            }
        }

    .. code-block:: python

        import qsharp
        from Sample import RunMain

        with qsharp.capture_diagnostics() as diagnostics:
            RunMain.simulate()

        print(len(diagnostics)) # will print 1

    :param passthrough: If `True`, captured diagnostics will also be displayed
        as normal. By default, diagnostic outputs captured by this context
        manager will not be displayed.
    :param as_qobj: If `True`, this context manager will attempt to convert
        captured diagnostics representing quantum states and operations into
        QuTiP objects. This option requires that QuTiP is installed and
        can be imported.
    """
    # Before proceeding, check that if we were asked to convert to qobj data
    # that we can actually import qutip.
    if as_qobj:
        # We don't actually need QuTiP here, but are only importing to capture
        # exceptions as early as possible so as to provide actionable error
        # messages to the user.
        qt = try_import_qutip()
        if isinstance(qt, ImportFailure):
            raise ImportError("as_qobj was set to `True`, but cannot convert captured diagnostics to QObj since QuTiP failed to import.") from qt.cause

        from qsharp.qobj import convert_diagnostic_to_qobj

    processed_data = []
    with client.capture_diagnostics(passthrough=passthrough) as data:
        yield processed_data

        # Apply any postprocessing needed here and append to processed_data.
        for diagnostic in data:
            if as_qobj:
                converted = convert_diagnostic_to_qobj(diagnostic)
                if converted is not None:
                    diagnostic = converted
            processed_data.append(diagnostic)

## STARTUP ##

client = _start_client()
config = Config(client)
packages = Packages(client)
projects = Projects(client)
_experimental_versions = None

# Make sure that we're last on the meta_path so that actual modules are loaded
# first.
sys.meta_path.append(QSharpModuleFinder())

# If using IPython, forward some useful IQ# magic commands as IPython magic
# commands and define a couple new magic commands for IPython.
try:
    if __IPYTHON__:
        import qsharp.ipython_magic
        qsharp.ipython_magic.register_magics()
except NameError:
    pass

# Needed to recognize PEP 420 packages as subpackages.
import pkg_resources
pkg_resources.declare_namespace(__name__)